"""
In this program we try to derive some results of the following paper:
Machine Learning for Quantum Mechanics in a Nutshell by Matthias Rupp 
  (https://doi.org/10.1002/qua.24954) 

To better understand I recomend the follwing paper:
Fast and Accurate Modeling of Molecular Atomization Energies with Machine Learning 
  by Matthias Rupp et. al (https://doi.org/10.1103/PhysRevLett.108.058301)

The aim is to estimate atomization energies of small organic molecules from machine learning.

Mojtaba Alaei 2018 
"""


import numpy as np
import numpy.linalg as la
import scipy.constants as const
from scipy.linalg import  cho_solve, cho_factor
import matplotlib.pyplot as plt

#step 0
####### Molecule CLASS#########
class Molecule:
     """
     Molecule class contains following properties a molecule:
     n:    number of atoms in the molecule
     ind:  index of molecule
     en:   atomization energy of the molecule
     atom: type of atoms in the molecule
     pos:  position (force field and DFT) of atoms in the molecule 
     nonh: number of non-H atoms in the molecule
     """
     def __init__(self,n,ind, en,atom, pos,nonh):
          self.n=n
          self.ind=ind
          self.en=en
          self.atom=atom
          self.pos=pos
          self.nonh=nonh
     def __lt__(self, other):
         return self.n < other.n
####### End of Molecule CLASS#########
     
     

#step 1
#######Read data from "dsgdb7ae2.xyz" and save it to a list (molecule)#########

molecule=[]   # A list to save properties of each molecule
dataxyz=open("dsgdb7ae2.xyz","r")

while True:
    line = dataxyz.readline()
    if not line:
         break
    pos=[]
    atom=[]
    words=line.split()
    if len(words)==1:
       n=int(words[0])
       line = dataxyz.readline()
       words=line.split()
       ind=int(words[0]) 
       ae=float(words[1]) 
       for i in range(n):
          line = dataxyz.readline()
          words=line.split()
          atom.append(words[0])
          pos.append(words[1:])
    m=Molecule(n,ind,ae,atom,pos,1)
    molecule.append(m)
dataxyz.close()
#######End of Read data and save data  ########################################


#step 1_0
#####Check number of molecules (7102) and correct index of some molecules!#####
N=len(molecule)
print("Number of molecule=", N)
#some index of molecules are not correct (for example  there is a wrong index for 89th molecule) 
for j in range(N):
   # correct index of molecules (re-indexing)
   molecule[j].ind=j
################################################################################
      


#step 1_1
########## set the number of non-H atoms of the molecules##################################
for j in range(N):
   nh=0
   n=molecule[j].n
   for i in range(n):
      if molecule[j].atom[i]=='H':
           nh=nh+1
   molecule[j].nonh=molecule[j].n-nh

#backup the molecules list
molecule_backup=molecule.copy()
###############################################################################


###### Create table 3 of the paper (distribution of molecular size)

#enumeration of the molecules according to their non-H atoms
n1=0;n2=0;n3=0;n4=0;n5=0;n6=0;n7=0
for j in range(N):
    if (molecule[j].nonh) == 1:
        n1=n1+1     
    if (molecule[j].nonh) == 2:
        n2=n2+1     
    if (molecule[j].nonh) == 3:
        n3=n3+1     
    if (molecule[j].nonh) == 4:
        n4=n4+1     
    if (molecule[j].nonh) == 5:
        n5=n5+1     
    if (molecule[j].nonh) == 6:
        n6=n6+1     
    if (molecule[j].nonh) == 7:
        n7=n7+1     
print("Table 3:")
print("non-H atoms", " 1","2","3"," 4"," 5","  6","  7")
print("n moulecules", n1,n2,n3,n4,n5,n6,n7)
####################################################


#step 3
##############Creating a training set##########################
molecule_t=[]
ind_train=[]   # a list to save index of training set

#first we select all of molecules which have 4 to 1 non-H atoms
for j in range(N):
    if molecule[j].nonh <= 4: 
         molecule_t.append(molecule[j])
         ind_train.append(molecule[j].ind)      

#remove these molecule from the main list
k=len(molecule_t)
for j in range(k):
     molecule.remove(molecule_t[j])


# sort remaining molecules by number of atoms:
molecule.sort() 

#select molecules by a step=(N-k)/(1000-k) ~8 
step=(N-k)/(1000-k)
select_list=np.round(np.arange(0,N-k,step))
select_list=select_list.astype(int)
for j in select_list:
    molecule_t.append(molecule[j])
    ind_train.append(molecule[j].ind)

#remove these molecule from the main list
k1=len(molecule_t)
for j in range(k,k1):
    molecule.remove(molecule_t[j])

# save remaining molecules to a new list 
molecule_out=molecule.copy()

ind=list(range(N))
for j in ind_train:
   ind.remove(j)
ind_predict=ind      # a list of remaining molecule as the prediction set
######################################################

#step 4
################Creat some arrays: 
### xyz to save positions, 
### Z to save atomic number,
### CM to save Coulomb matrix
del molecule

molecule=molecule_backup.copy()

natoms=[]
for i in range(N):
  natoms.append(molecule[i].n)
nmax=max(natoms) # the maximum number of atoms in a molecule
del natoms
xyz=np.zeros([N,nmax,3])
Z=np.zeros([N,nmax])
CM=np.zeros([N,nmax,nmax])
################################


#step 5
###############Construct Coulomb matrix  #####################
def atomicnumber(name):
   """
   convert atom name to its atomic number
   """
   if name=='H':
     return 1
   if name=='C':
     return 6
   if name=='N':
     return 7
   if name=='O':
     return 8
   if name=='S':
     return 16



for j in range(N):
   for i in range(molecule[j].n):
       Z[j,i]=atomicnumber(molecule[j].atom[i])
       xyz[j,i,0]=float(molecule[j].pos[i][0])
       xyz[j,i,1]=float(molecule[j].pos[i][1])
       xyz[j,i,2]=float(molecule[j].pos[i][2])
       #xyz[j,i,0]=float(molecule[j].pos[i][3])   #dft positions
       #xyz[j,i,1]=float(molecule[j].pos[i][4])   #dft positions
       #xyz[j,i,2]=float(molecule[j].pos[i][5])   #dft positions
ang2bohr=1.8897261
xyz=xyz*ang2bohr
for j in range(N):
   for i in range(molecule[j].n):
      for k in range(molecule[j].n):
          if i==k:
             CM[j,i,k]=0.5*Z[j,i]**2.4
          else:
             CM[j,i,k]=Z[j,i]*Z[j,k]/la.norm(xyz[j,i,:]-xyz[j,k,:])

#Sort Coulomb matrix by its row norm. Please see: 
#https://stackoverflow.com/questions/36885418/sort-symmetric-numpy-2d-array-by-function-norm 
#https://coderwall.com/p/uhskuq/reverse-array-in-python
for j in range(N):
    CMj=CM[j,:,:]
    indexlist = np.argsort(la.norm(CMj,axis=1))
    indexlist_r=indexlist[::-1]
    sym_sortedCM = CMj[indexlist_r][:,indexlist_r]
    CM[j,:,:]=sym_sortedCM
########################################################################


#step 6
############Convert  Coulomb matrix to  a vector #####################
# convert the CM to upper triangle and then exctract all of upper triangle matrix to a vector 
# in the Mathias paper, he covert CM to lower triangle and .. but both method do the same because CM is a symmetric matrix


I=[]
J=[]
for i in range(nmax):
   for j in range(i+1):
     I.append(i)
     J.append(j)


indices=(I,J)

CM_j=CM[0,:,:]
CM_vec=CM_j[indices]
cm_vec_size=CM_vec.size # (23+22+21+20+...+1=276 or nmax*(nmax+1)/2)
CM_vec=np.zeros([N,cm_vec_size])

for j in range(N):
    CM_j=CM[j,:,:]
    CM_vec[j,:]=CM_j[indices]

#step7
############Construct K matrix #########################
# Calculate the pairwise distances between Coulomb vectors
CM_vec_d=np.zeros([N,N])
for j in range(N):
    for k in range(j):
        CM_vec_d[j,k]=la.norm(CM_vec[j,:]-CM_vec[k,:])
        CM_vec_d[k,j]=CM_vec_d[j,k]


# hyperparmeters:  {\Lambda, \Sigma} = Power[2., {-22, 9.5}]; form mathematica source of the paper

lambda_p=2.0**(-22)  # is about 2.38418e-07 ~ 10e-6.5
sigma_p=2.0**(9.5)   # is about 724.07734
Ntrain=len(ind_train)
Kmatrix=np.zeros([Ntrain,Ntrain])


####TO COMPLETE#############

######################################################


#step 8 
############### solve equation (K+\lambda I) \alpha = y  by Cholesky decomposition#####

#####TO COMPLETE#################

#alpha=cho_solve(cho_factor(....), ...)   
#####################################################################################


#step 9
##################prediction#########################################################

######TO COMPLETE#######################

################################################################################
